@@ -7,7 +7,11 @@
 #include "commands/search/search-packages-hinthack.h"
 
 #include <zypp/base/Algorithm.h>
+#include <zypp/sat/Solvable.h>
+#include <zypp/Capability.h>
+#include <zypp/PoolQueryResult.h>
 
+#include <unordered_map>
 
 namespace zypp
 {
@@ -22,6 +26,15 @@ namespace zypp
             }
       );
     }
+
+    Value setSolvAttrOptional ( boost::optional<zypp::sat::SolvAttr> &target_r, zypp::sat::SolvAttr value_r ) {
+      return Value (
+          noDefaultValue,
+          [ &target_r, value_r ] ( const CommandOption &, const boost::optional<std::string> & ) {
+            target_r = value_r;
+          }
+        );
+    }
   }
 }
 
@@ -33,28 +46,18 @@ namespace
     for ( const auto & pi : ResPool::instance().byName( name_r ) )
     {
       if ( Edition::match( pi.edition(), edition_r ) == 0 )
-	return true;
+        return true;
     }
     return false;
   }
 }
 
 
-SearchCmd::SearchCmd(std::vector<std::string> &&commandAliases_r, CmdMode cmdMode_r  ) :
-  ZypperBaseCommand (
-    std::move( commandAliases_r ),
-    std::string(),
-    std::string(),
-    std::string(),
-    ResetRepoManager
-  ), _cmdMode ( cmdMode_r )
+SearchCmd::SearchCmd( std::vector<std::string> &&commandAliases_r )
+: ZypperBaseCommand( std::move( commandAliases_r ), std::string(), std::string(), std::string(), ResetRepoManager )
 {
-  if ( _cmdMode == CmdMode::Search ) {
-    this->addOptionSet( _initReposOpts );
-    this->addOptionSet( _notInstalledOpts );
-  }
-  _sortOpts.setCompatibilityMode( CompatModeBits::EnableNewOpt | CompatModeBits::EnableNewOpt );
-  _initReposOpts.setCompatibilityMode( CompatModeBits::EnableNewOpt | CompatModeBits::EnableNewOpt );
+  _sortOpts.setCompatibilityMode( CompatModeBits::EnableNewOpt );
+  _initReposOpts.setCompatibilityMode( CompatModeBits::EnableNewOpt );
 }
 
 void SearchCmd::setMode(const MatchMode &mode_r)
@@ -75,31 +78,20 @@ std::string SearchCmd::summary() const
 
 std::vector<std::string> SearchCmd::synopsis() const
 {
-  if ( _cmdMode == CmdMode::Search ) {
-    return {
-      // translators: command synopsis; do not translate lowercase words
-      _("search (se) [OPTIONS] [QUERYSTRING] ...")
-    };
-  }
   return {
     // translators: command synopsis; do not translate lowercase words
-    _("patch-search [OPTIONS] [QUERYSTRING...]")
+    _("search (se) [OPTIONS] [QUERYSTRING] ...")
   };
 }
 
 std::string SearchCmd::description() const
 {
-  if ( _cmdMode == CmdMode::Search ) {
-    return (
-      // translators: command description
-      std::string( _("Search for packages matching any of the given search strings.") ) + "\n\n"
-      // translators: command description
-      + _("* and ? wildcards can also be used within search strings. If a search string is enclosed in '/', it's interpreted as a regular expression.")
-    );
-  }
-
-  // translators: command description
-  return _("Search for patches matching given search strings.");
+  return (
+    // translators: command description
+    std::string( _("Search for packages matching any of the given search strings.") ) + "\n\n"
+    // translators: command description
+    + _("* and ? wildcards can also be used within search strings. If a search string is enclosed in '/', it's interpreted as a regular expression.")
+  );
 }
 
 zypp::ZyppFlags::CommandGroup SearchCmd::cmdOptions() const
@@ -119,106 +111,110 @@ zypp::ZyppFlags::CommandGroup SearchCmd::cmdOptions() const
       { "match-exact", 'x', ZyppFlags::NoArgument, ZyppFlags::WriteFixedValueType( that._mode, MatchMode::Exact ),
             // translators: -x, --match-exact
             _("Searches for an exact match of the search strings.")
-      }
-    },{
-      { "match-substrings", "match-words", "match-exact" }
-    }
-  };
-
-  if ( _cmdMode == CmdMode::Search ) {
-    grp.options.insert( grp.options.end(),
-    {
-        { "provides", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::provides ),
-              // translators: --provides
-              _("Search for packages which provide the search strings.")
-        },
-        { "requires", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::requires ),
-              // translators: --requires
-              _("Search for packages which require the search strings.")
-        },
-        { "recommends", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::recommends ),
-              // translators: --recommends
-              _("Search for packages which recommend the search strings.")
-        },
-        { "supplements", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::supplements ),
-              // translators: --supplements
-              _("Search for packages which supplement the search strings.")
-        },
-        { "conflicts", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::conflicts ),
-              // translators: --conflicts
-              _("Search packages conflicting with search strings.")
-        },
-        { "obsoletes", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::obsoletes ),
-              // translators: --obsoletes
-              _("Search for packages which obsolete the search strings.")
-        },
-        { "suggests", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::suggests ),
-              // translators: --suggests
-              _("Search for packages which suggest the search strings.")
-        },
-        CommonFlags::resKindSetFlag( that._requestedTypes,
-              // translators: -t, --type <TYPE>
-              _("Search only for packages of the specified type.")
-        ),
-        { "name", 'n', ZyppFlags::NoArgument, ZyppFlags::BoolType( &that._forceNameAttr, ZyppFlags::StoreTrue ),
-              // translators: -n, --name
-             _("Useful together with dependency options, otherwise searching in package name is default.")
-        },
-        { "file-list", 'f', ZyppFlags::NoArgument, ZyppFlags::BoolType( &that._searchFileList, ZyppFlags::StoreTrue, _searchFileList ),
-              // translators: -f, --file-list
-              _("Search for a match in the file list of packages.")
-        }
-    } );
-  }
-
-  grp.options.insert( grp.options.end(),
-  {
+      },
+      { "provides", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::dep_provides ),
+        // translators: --provides
+        _("Search for packages which provide the search strings.")
+      },
+      { "requires", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::dep_requires ),
+        // translators: --requires
+        _("Search for packages which require the search strings.")
+      },
+      { "recommends", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::dep_recommends ),
+        // translators: --recommends
+        _("Search for packages which recommend the search strings.")
+      },
+      { "supplements", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::dep_supplements ),
+        // translators: --supplements
+        _("Search for packages which supplement the search strings.")
+      },
+      { "conflicts", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::dep_conflicts ),
+        // translators: --conflicts
+        _("Search packages conflicting with search strings.")
+      },
+      { "obsoletes", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::dep_obsoletes ),
+        // translators: --obsoletes
+        _("Search for packages which obsolete the search strings.")
+      },
+      { "suggests", '\0', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::dep_suggests ),
+        // translators: --suggests
+        _("Search for packages which suggest the search strings.")
+      },
+      { "provides-pkg", '\0', ZyppFlags::NoArgument, ZyppFlags::setSolvAttrOptional( that._requestedReverseSearch, sat::SolvAttr::dep_provides ),
+        // translators: --provides-pkg
+        _("Search for all packages that provide any of the provides of the package(s) matched by the input parameters.")
+      },
+      { "requires-pkg", '\0', ZyppFlags::NoArgument, ZyppFlags::setSolvAttrOptional( that._requestedReverseSearch, sat::SolvAttr::dep_requires ),
+        // translators: --requires-pkg
+        _("Search for all packages that require any of the provides of the package(s) matched by the input parameters.")
+      },
+      { "recommends-pkg", '\0', ZyppFlags::NoArgument, ZyppFlags::setSolvAttrOptional( that._requestedReverseSearch, sat::SolvAttr::dep_recommends ),
+        // translators: --recommends-pkg
+        _("Search for all packages that recommend any of the provides of the package(s) matched by the input parameters.")
+      },
+      { "supplements-pkg", '\0', ZyppFlags::NoArgument, ZyppFlags::setSolvAttrOptional( that._requestedReverseSearch, sat::SolvAttr::dep_supplements ),
+        // translators: --supplements-pkg
+        _("Search for all packages that supplement any of the provides of the package(s) matched by the input parameters.")
+      },
+      { "conflicts-pkg", '\0', ZyppFlags::NoArgument, ZyppFlags::setSolvAttrOptional( that._requestedReverseSearch, sat::SolvAttr::dep_conflicts ),
+        // translators: --conflicts-pkg
+        _("Search for all packages that conflict with any of the package(s) matched by the input parameters.")
+      },
+      { "obsoletes-pkg", '\0', ZyppFlags::NoArgument, ZyppFlags::setSolvAttrOptional( that._requestedReverseSearch, sat::SolvAttr::dep_obsoletes ),
+        // translators: --obsoletes-pkg
+        _("Search for all packages that obsolete any of the package(s) matched by the input parameters.")
+      },
+      { "suggests-pkg", '\0', ZyppFlags::NoArgument, ZyppFlags::setSolvAttrOptional( that._requestedReverseSearch, sat::SolvAttr::dep_suggests ),
+        // translators: --suggests-pkg
+        _("Search for all packages that suggest any of the provides of the package(s) matched by the input parameters.")
+      },
+      CommonFlags::resKindSetFlag( that._requestedTypes,
+                                   // translators: -t, --type <TYPE>
+                                   _("Search only for packages of the specified type.")
+      ),
+      { "name", 'n', ZyppFlags::NoArgument, ZyppFlags::BoolType( &that._forceNameAttr, ZyppFlags::StoreTrue ),
+        // translators: -n, --name
+        _("Useful together with dependency options, otherwise searching in package name is default.")
+      },
+      { "file-list", 'f', ZyppFlags::NoArgument, ZyppFlags::appendSolvAttrToSet( that._requestedDeps, sat::SolvAttr::filelist ),
+        // translators: -f, --file-list
+        _("Search for a match in the file list of packages.")
+      },
       { "search-descriptions", 'd', ZyppFlags::NoArgument, ZyppFlags::BoolType( &that._searchDesc, ZyppFlags::StoreTrue, _searchDesc ),
-            // translators: -d, --search-descriptions
-            _("Search also in package summaries and descriptions.")
+        // translators: -d, --search-descriptions
+        _("Search also in package summaries and descriptions.")
       },
       {"case-sensitive", 'C', ZyppFlags::NoArgument, ZyppFlags::BoolType( &that._caseSensitive, ZyppFlags::StoreTrue, _caseSensitive ),
-            // translators: -C, --case-sensitive
-            _("Perform case-sensitive search.")
-      }
-  });
-
-  if ( _cmdMode == CmdMode::Search ) {
-    grp.options.insert( grp.options.end(),
-    {
+        // translators: -C, --case-sensitive
+        _("Perform case-sensitive search.")
+      },
       CommonFlags::detailsFlag( that._details, 's',
-            // translators: -s, --details
-            _("Show each available version in each repository on a separate line.")
+                                // translators: -s, --details
+                                _("Show each available version in each repository on a separate line.")
       ),
       {"verbose", 'v', ZyppFlags::NoArgument, ZyppFlags::BoolType( &that._verbose, ZyppFlags::StoreTrue, _caseSensitive ),
-            // translators: -v, --verbose
-            _("Like --details, with additional information where the search has matched (useful for search in dependencies).")
+        // translators: -v, --verbose
+        _("Like --details, with additional information where the search has matched (useful for search in dependencies).")
       }
-    });
-  }
+    },
+    {
+      { "match-substrings", "match-words", "match-exact" },
+      { "provides-pkg", "requires-pkg", "recommends-pkg", "supplements-pkg", "conflicts-pkg", "obsoletes-pkg", "suggests-pkg"  }
+    }
+  };
 
   return grp;
 }
 
 std::string SearchCmd::help()
 {
-  if ( _cmdMode == CmdMode::Search )
-    return ZypperBaseCommand::help();
-
-  CommandHelpFormater cliHlp;
-  for ( const std::string &syn : synopsis() )
-    cliHlp.synopsis( syn );
-
-  cliHlp.description( description())
-      .descriptionAliasCmd( "zypper -r search -t patch --detail ..." );
-  return cliHlp;
+  return ZypperBaseCommand::help();
 }
 
 void SearchCmd::doReset()
 {
   _mode = MatchMode::Default;
   _forceNameAttr = false;
-  _searchFileList = false;
   _searchDesc = false;
   _caseSensitive = false;
   _details = false;
@@ -243,6 +239,13 @@ int SearchCmd::execute( Zypper &zypper, const std::vector<std::string> &position
     inst_notinst = true;
     zypper.configNoConst().no_refresh = true;
     //  query.setInstalledOnly();
+    // NOTE Not sure why setInstalledOnly was disabled. Anyway, if you consider
+    // to re-enable it, check (at least) FillSearchTableSolvable. The code there
+    // discards an installed item if an identicalAvailable is present, expecting
+    // that the query will deliver this too (even if inst_notinst == true).
+    // Maybe done to print the correct repo (rather than @System), but if that's
+    // all, the overhead of processing all available (and discarding most of them)
+    // is bigger than chacking for an identicalAvailable when printing the installed.
   }
 
   switch ( _mode  ) {
@@ -262,9 +265,7 @@ int SearchCmd::execute( Zypper &zypper, const std::vector<std::string> &position
   if ( _caseSensitive )
     query.setCaseSensitive();
 
-  if ( _cmdMode == CmdMode::RugPatchSearch )
-    query.addKind( ResKind::patch );
-  else if ( _requestedTypes.size() > 0 )
+  if ( _requestedTypes.size() > 0 )
   {
     for ( const ResKind &knd : _requestedTypes )
       query.addKind( knd );
@@ -337,7 +338,7 @@ int SearchCmd::execute( Zypper &zypper, const std::vector<std::string> &position
       query.addDependency( attr , name, cap.detail().op(), cap.detail().ed(), Arch(cap.detail().arch()), matchmode );
 
       //handle special cases
-      if ( attr == sat::SolvAttr::provides && str::regex_match( name.c_str(), std::string("^/") ) ) {
+      if ( attr == sat::SolvAttr::dep_provides && str::regex_match( name.c_str(), std::string("^/") ) ) {
         // in case of path names also search in file list
         query.setFilesMatchFullPath( true );
         query.addDependency( sat::SolvAttr::filelist , name, cap.detail().op(), cap.detail().ed(), Arch(cap.detail().arch()), matchmode );
@@ -385,55 +386,90 @@ int SearchCmd::execute( Zypper &zypper, const std::vector<std::string> &position
     }
   }
 
-  // Output query result...
   Table t;
   try
   {
-    if ( _cmdMode == CmdMode::RugPatchSearch )
-    {
-      FillPatchesTable callback( t, inst_notinst );
-      invokeOnEach( query.poolItemBegin(), query.poolItemEnd(), callback );
-    }
-    else if ( details )
-    {
-      FillSearchTableSolvable callback( t, inst_notinst );
-      if ( _verbose )
+    if ( _requestedReverseSearch.is_initialized() ) {
+
+      std::unordered_map< sat::Solvable, CapabilitySet > matchedSolvables;
+      const auto reqSearchAttrib = _requestedReverseSearch.get();
+
+      for ( const auto slv : query ) {
+
+        bool isInstalled = slv.isSystem();
+        if ( isInstalled && _notInstalledOpts._mode == SolvableFilterMode::ShowOnlyNotInstalled )
+          continue;
+        if ( !isInstalled && _notInstalledOpts._mode == SolvableFilterMode::ShowOnlyInstalled )
+          continue;
+
+        sat::Queue q = sat::Pool::instance().whatMatchesSolvable( reqSearchAttrib, slv  );
+
+        for ( auto matchedSolvId : q ) {
+
+          sat::Solvable matchedSolv ( static_cast<sat::Solvable::IdType>(matchedSolvId) );
+          auto p = matchedSolvables.insert( make_pair( std::move(matchedSolv), CapabilitySet()) );
+
+          if ( _verbose ) {
+            CapabilitySet matchedCaps = matchedSolv.matchesSolvable( reqSearchAttrib, slv).second;
+            p.first->second.insert( matchedCaps.begin(), matchedCaps.end() );
+          }
+        }
+      }
+
+      if ( details ) {
+        FillSearchTableSolvable callback( t, inst_notinst );
+        std::for_each( matchedSolvables.begin(), matchedSolvables.end(), [&callback, verb = _verbose, &reqSearchAttrib ]( auto elem ){
+          if ( verb )
+            callback( elem.first, reqSearchAttrib, elem.second );
+          else
+            callback( elem.first, reqSearchAttrib, {} );
+        } );
+      } else {
+
+        PoolQueryResult res;
+        std::for_each( matchedSolvables.begin(), matchedSolvables.end(), [ &res ]( const auto &v ){ res+=v.first; } );
+
+        FillSearchTableSelectable callback( t, inst_notinst );
+        std::for_each( res.selectableBegin(), res.selectableEnd(), callback);
+      }
+
+    } else {
+      if ( details )
       {
-        // Option 'verbose' shows where (e.g. in 'requires', 'name') the search has matched.
-        // Info is available from PoolQuery::const_iterator.
-        for_( it, query.begin(), query.end() )
-          callback( it );
+        FillSearchTableSolvable callback( t, inst_notinst );
+        if ( _verbose )
+        {
+          // Option 'verbose' shows where (e.g. in 'requires', 'name') the search has matched.
+          // Info is available from PoolQuery::const_iterator.
+          for_( it, query.begin(), query.end() )
+            callback( it );
+        }
+        else
+        {
+          for ( const auto slv : query )
+            callback( slv );
+        }
       }
       else
       {
-        for ( const auto slv : query )
-          callback( slv );
+        FillSearchTableSelectable callback( t, inst_notinst );
+        invokeOnEach( query.selectableBegin(), query.selectableEnd(), callback );
       }
     }
-    else
-    {
-      FillSearchTableSelectable callback( t, inst_notinst );
-      invokeOnEach( query.selectableBegin(), query.selectableEnd(), callback );
-    }
 
     if ( t.empty() )
     {
       // translators: empty search result message
       zypper.out().info(_("No matching items found."), Out::QUIET );
-      zypper.setExitCode( ZYPPER_EXIT_INF_CAP_NOT_FOUND );
+      if ( !zypper.config().ignore_unknown ) {
+        zypper.setExitInfoCode( ZYPPER_EXIT_INF_CAP_NOT_FOUND );
+      }
     }
     else
     {
       cout << endl; //! \todo  out().separator()?
 
-      if ( _cmdMode == CmdMode::RugPatchSearch )
-      {
-        if ( _sortOpts._mode == SortResultOptionSet::ByRepo )
-          t.sort( { 0, 1, Table::UserData } );
-        else
-          t.sort( { 1, Table::UserData } ); // sort by name
-      }
-      else if ( _details )
+      if ( _details )
       {
         if ( _sortOpts._mode == SortResultOptionSet::ByRepo )
           t.sort( { 5, 1, Table::UserData } );
@@ -451,15 +487,16 @@ int SearchCmd::execute( Zypper &zypper, const std::vector<std::string> &position
       //cout << t; //! \todo out().table()?
       zypper.out().searchResult( t );
     }
-  }
-  catch ( const Exception & e )
-  {
+
+    if ( !_requestedReverseSearch.is_initialized() )
+      searchPackagesHintHack::callOrNotify( zypper );
+
+  } catch ( const Exception & e )  {
     zypper.out().error( e, _("Problem occurred initializing or executing the search query") + std::string(":"),
-                 std::string(_("See the above message for a hint.")) + " "
-                 + _("Running 'zypper refresh' as root might resolve the problem.") );
+      std::string(_("See the above message for a hint.")) + " "
+        + _("Running 'zypper refresh' as root might resolve the problem.") );
     zypper.setExitCode( ZYPPER_EXIT_ERR_ZYPP );
   }
 
-  searchPackagesHintHack::callOrNotify( zypper );
   return zypper.exitCode();
 }
